/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "libzip-private.h"

#include <stdlib.h>

#if ENABLE_DEFLATE
#include <zlib.h>
#endif

#if ENABLE_BZIP2
#include <bzlib.h>
#endif

#if ENABLE_PPMD
#include "ppmdi1/PPMd.h"
#endif

/* ===== internal data structures ===== */

typedef size_t
(*libzip_decoder_read_t)(libzip_decoder_t *decoder,
                         void *data, size_t length);

struct libzip_decoder_t {
    libzip_stream_t *stream;
    libzip_stream_t *decoder_created_stream;
    bool initialized;
    libzip_compression_t compression;

#if ENABLE_DEFLATE || ENABLE_BZIP2 || ENABLE_PPMD
    union {
#if ENABLE_DEFLATE
        struct {
            uint8_t buffer[65536];
            z_stream stream;
            int finish;
        } inflate;
#endif

#if ENABLE_BZIP2
        struct {
            char buffer[65536];
            bz_stream stream;
        } bzip2;
#endif

#if ENABLE_PPMD
        struct {
            unsigned char inbuf[65536];
            unsigned char *inptr;
            size_t inlen;

            unsigned char *outptr;
            size_t outlen;

            int eof;

//            FILE *file;
//            int incnt;
//            int outcnt;
        } ppmd;
#endif
    } x;
#endif

    libzip_decoder_read_t read;
};

/* ===== libzip_store_decoder ===== */

static size_t
libzip_store_decoder_read(libzip_decoder_t *decoder, 
                          void *data, size_t length)
{
    return_val_if_fail(NULL != decoder, 0);
    return libzip_stream_read(decoder->stream, data, length);
}

/* ===== libzip_inflate_decoder ===== */

#if ENABLE_DEFLATE

static size_t
libzip_inflate_decoder_read(libzip_decoder_t *decoder,
                            void *data, size_t length)
{
    size_t bytes_read = 0;
    int result = Z_OK;
    z_stream *stream;

    return_val_if_fail(NULL != decoder, 0);
    return_val_if_fail(NULL != data, 0);

    stream = &decoder->x.inflate.stream;

    if (!decoder->initialized) {
        stream->zalloc = Z_NULL;
        stream->zfree = Z_NULL;
        stream->opaque = Z_NULL;

        /* a negative value for windowBits causes the stream
         * to _not_ expect a header */
        result = inflateInit2(stream, -15);
        debugf("inflateInit2: %s (%d)", zError(result), result);
        decoder->initialized = (Z_OK == result);
    }

    if (decoder->initialized) {
        stream->avail_out = length;
        stream->next_out = data;

        while(stream->avail_out) {
            if (0 == stream->avail_in) {
                stream->avail_in = libzip_stream_read(
                    decoder->stream, decoder->x.inflate.buffer,
                    sizeof(decoder->x.inflate.buffer));
                stream->next_in = decoder->x.inflate.buffer;
            }

            result = inflate(stream, decoder->x.inflate.finish);
            debugf("inflate: %s (%d)", zError(result), result);
        }

        bytes_read += (length - stream->avail_out);

        if (0 == bytes_read && Z_NO_FLUSH == decoder->x.inflate.finish) {
            decoder->x.inflate.finish = Z_FINISH;
            return libzip_inflate_decoder_read(decoder, data, length);
        }

        if (decoder->x.inflate.finish) {
            result = inflateEnd(&decoder->x.inflate.stream);
            debugf("inflateEnd: %s (%d)", zError(result), result);
            decoder->initialized = false;
        }
    }

    return bytes_read;
}

#endif

/* ===== libzip_decoder_bzip2 ===== */

#if ENABLE_BZIP2

static size_t
libzip_bzip2_decoder_read(libzip_decoder_t *decoder,
                          void *data, size_t length)
{
    size_t bytes_read = 0;
    int result = BZ_OK;
    bz_stream *stream;

    return_val_if_fail(NULL != decoder, 0);
    return_val_if_fail(NULL != data, 0);

    stream = &decoder->x.bzip2.stream;

    if (!decoder->initialized) {
        stream->bzalloc = Z_NULL;
        stream->bzfree = Z_NULL;
        stream->opaque = Z_NULL;

        result = BZ2_bzDecompressInit(stream, 0, 0);
        debugf("BZ2_bzDecompressInit: %d", result);
        decoder->initialized = (BZ_OK == result);
    }

    if (decoder->initialized) {
        stream->avail_out = length;
        stream->next_out = data;

        while(stream->avail_out && BZ_STREAM_END != result) {
            if (0 == stream->avail_in) {
                stream->avail_in = libzip_stream_read(decoder->stream,
                    decoder->x.bzip2.buffer, sizeof(decoder->x.bzip2.buffer));
                stream->next_in = decoder->x.bzip2.buffer;
            }

            result = BZ2_bzDecompress(stream);
            debugf("BZ2_bzDecompress: %d", result);
        }

        bytes_read += (length - stream->avail_out);

        if (BZ_STREAM_END == result) {
            result = BZ2_bzDecompressEnd(&decoder->x.bzip2.stream);
            debugf("BZ2_bzDecompressEnd: %d", result);
            decoder->initialized = false;
        }
    }

    return bytes_read;
}

#endif
            
/* ===== libzip_ppmd_decoder ===== */

#if ENABLE_PPMD

void _STDCALL 
PrintInfo(_PPMD_FILE* DecodedFile __attribute__((unused)),
          _PPMD_FILE* EncodedFile __attribute__((unused)))
{
}

static size_t
libzip_ppmd_decoder_read(libzip_decoder_t *decoder,
                         void *data, size_t length)
{
    size_t bytes_read = 0;
    uint16_t params;
    size_t len;

    return_val_if_fail(NULL != decoder, 0);
    return_val_if_fail(NULL != data, 0);

    if (decoder->x.ppmd.eof)
        return 0;

    decoder->x.ppmd.outptr = data;
    decoder->x.ppmd.outlen = length;

    if (!decoder->initialized) {
        len = libzip_stream_read(decoder->stream, &params, sizeof params);

        if (len != sizeof params) {
            error("Corrupted PPMd header");
            return 0;
        }

		{
			MR_METHOD mrm = (params >> 12) & 15;
			int suballoc_size = ((params >> 4) & 255) + 1;
			int max_order = (params & 15) + 1;

			debugf("PPMd: model restoration method: %d\n", mrm);
			debugf("PPMd: sub allocator size: %d\n", suballoc_size);
			debugf("PPMd: maximum model order: %d\n", max_order);

			if (!StartSubAllocator(suballoc_size)) {
				errorf("Out of memory (PPMd sub-allocator; size=%d)", suballoc_size);
				return 0;
			}

			decoder->initialized = TRUE;
			StartDecode((_PPMD_FILE*)decoder, max_order, mrm);
		}

//        static int i = 0;
//        char filename[100];
//        snprintf(filename, sizeof filename, "ppmd%d", i++);
//decoder->x.ppmd.incnt = decoder->x.ppmd.outcnt = 0;
//        decoder->x.ppmd.file = fopen(filename, "wb");
//        DecodeFile((_PPMD_FILE*)decoder, (_PPMD_FILE*)decoder, max_order, mrm);
//        fflush(decoder->x.ppmd.file);
//        fclose(decoder->x.ppmd.file);
    }

    if (decoder->initialized)
        bytes_read = DecodeNext((_PPMD_FILE*)decoder, data, length, &decoder->x.ppmd.eof);

    if (decoder->x.ppmd.eof)
        StopSubAllocator();

    return bytes_read;
}

int
libzip_ppmd_decoder_getc(libzip_decoder_t *decoder)
{
    int c = -1;

    return_val_if_fail(NULL != decoder, -1);

    if (0 == decoder->x.ppmd.inlen) {
printf("%s: filling buffer\n", __FUNCTION__);
        decoder->x.ppmd.inptr = decoder->x.ppmd.inbuf;
        decoder->x.ppmd.inlen = libzip_stream_read(
            decoder->stream, decoder->x.ppmd.inptr,
            sizeof decoder->x.ppmd.inbuf);
    }

    if (decoder->x.ppmd.inlen > 0) {
        c = *decoder->x.ppmd.inptr++;
        --decoder->x.ppmd.inlen;
    }
//printf("%s: %d: %02x\n", __FUNCTION__, decoder->x.ppmd.incnt++, (unsigned char)c);

    return c;
}

int 
libzip_ppmd_decoder_putc(libzip_decoder_t *decoder, int c)
{
    return_val_if_fail(NULL != decoder, -1);

//printf("%s: %d: %02x\n", __FUNCTION__, decoder->x.ppmd.outcnt++, (unsigned char)c);

    if (decoder->x.ppmd.outlen) {
        *decoder->x.ppmd.outptr++ = c;
        --decoder->x.ppmd.outlen;

//        fputc(c, decoder->x.ppmd.file);
    }

    return c;
}

#endif
            
/* ===== libzip_default_decoder ===== */

static libzip_decoder_t*
libzip_simple_decoder_new(libzip_stream_t *stream,
                          libzip_decoder_read_t read)
{
    libzip_decoder_t* decoder = NULL;
    raise_if_fail(NULL != stream);

    decoder = libzip_new(libzip_decoder_t, 1);
    decoder->stream = stream;
    decoder->read = read;

finally:
    return decoder;
}

static libzip_decoder_t*
libzip_default_decoder_new(libzip_stream_t *stream,
                           libzip_compression_t compression)
{
    libzip_decoder_t* decoder = NULL;
    raise_if_fail(NULL != stream);

    switch(compression) {
        case LIBZIP_COMPRESSION_STORE:
            decoder = libzip_simple_decoder_new(stream, libzip_store_decoder_read);
            break;

#if ENABLE_DEFLATE
        case LIBZIP_COMPRESSION_DEFLATE:
            decoder = libzip_simple_decoder_new(stream, libzip_inflate_decoder_read);
            break;
#endif

#if ENABLE_BZIP2
        case LIBZIP_COMPRESSION_PKWARE_BZIP2:
            decoder = libzip_simple_decoder_new(stream, libzip_bzip2_decoder_read);
            break;
#endif

#if ENABLE_PPMD
        case LIBZIP_COMPRESSION_WINZIP_PPMD:
            decoder = libzip_simple_decoder_new(stream, libzip_ppmd_decoder_read);
            break;
#endif

        default:
            raisef("Unsupported compression method %d", compression);
            break;
    }
	
finally:
    return decoder;
}

/* ===== libzip_winzip_aes_decoder ===== */

static libzip_decoder_t*
libzip_winzip_aes_decoder_new(libzip_stream_t *stream, 
                              libzip_entry_t *entry,
                              const char *password)
{
    libzip_decoder_t *decoder = NULL;
    libzip_stream_t *encrypted_stream = NULL;
    libzip_winzip_aes_t *crypto_params = NULL;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != entry);
    raise_if_fail(NULL != password);
    raise_if_fail('\0' != password[0]);
 
    crypto_params = 
        (libzip_winzip_aes_t*) libzip_entry_find_extra(
        entry, LIBZIP_EXTRA_SIGNATURE_WINZIP_AES);
    raise_if_fail(NULL != crypto_params);

    encrypted_stream = libzip_aes_stream_open(
        stream, crypto_params->mode, password);
    raise_if_fail(NULL != encrypted_stream);

    decoder = libzip_default_decoder_new(
        encrypted_stream, crypto_params->compression);
    raise_if_fail(NULL != decoder);
    
    // Remember that we need to free this on decoder close
    decoder->decoder_created_stream = encrypted_stream;    

	free(crypto_params);

finally:
    if (encrypted_stream && !decoder)
        libzip_stream_close(encrypted_stream);

    return decoder;
}

/* ===== libzip_decoder ===== */

libzip_decoder_t*
libzip_decoder_new(libzip_stream_t *stream, 
                   libzip_entry_t *entry, 
                   const char *password)
{
    libzip_decoder_t* decoder = NULL;

    raise_if_fail(NULL != entry);
    raise_if_fail(NULL != stream);

    switch(entry->header.compression) {
        case LIBZIP_COMPRESSION_WINZIP_AES:
            decoder = libzip_winzip_aes_decoder_new(stream, entry, password);
            break;

        default:
            decoder = libzip_default_decoder_new(
                stream, entry->header.compression);
            break;
    }
    
    raise_if_fail(NULL != decoder);
    raise_if_fail(NULL != decoder->stream);
    raise_if_fail(NULL != decoder->read);

	decoder->compression = entry->header.compression;

finally:
    return decoder;
}

size_t
libzip_decoder_read(libzip_decoder_t *decoder,
                    void *data, size_t length)
{
    size_t bytes_written = 0;
    
    raise_if_fail(NULL != decoder);
    raise_if_fail(NULL != decoder->read);
    
    bytes_written += decoder->read(decoder, data, length);

finally:
    return bytes_written;
}

void
libzip_decoder_free(libzip_decoder_t *decoder)
{
    if (NULL != decoder)
    {
		if(decoder->compression == LIBZIP_COMPRESSION_DEFLATE)
		{
			// Tidy up inflation if we didn't inflate the whole file
			if(decoder->initialized)
				inflateEnd(&decoder->x.inflate.stream);
		}

		// Tidy up decryption stream
		if(NULL != decoder->decoder_created_stream)
			libzip_stream_close(decoder->decoder_created_stream);
			
        free(decoder);
	}
}

/* vim: set sw=4 sta et: */
