/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "libzip-private.h"
#include "fileencrypt/fileenc.h"

#include <stdlib.h>

#if ENABLE_DEFLATE
#include <zlib.h>
#else
#error TODO: Provide fallback implementation of CRC32 function.
#endif

#if ENABLE_BZIP2
#include <bzlib.h>
#endif

/* ===== internal data structures ===== */

struct libzip_encoder_t {
    libzip_stream_t *stream;
    uint16_t feature_level;
    uint16_t flags;

    uint8_t *next_out;
    uint32_t avail_out;
    uint32_t total_out;

    const uint8_t *next_in;
    uint32_t avail_in;
    uint32_t total_in;

    uint32_t checksum;

#if ENABLE_DEFLATE || ENABLE_BZIP2
    union {
#if ENABLE_DEFLATE
        struct {
            z_stream compressor;
            bool initialized;
        } deflate;
#endif

#if ENABLE_BZIP2
        struct {
            bz_stream compressor;
            bool initialized;
        } bzip2;
#endif
    } x;
#endif        

    bool (*write)(libzip_encoder_t *encoder);
    bool (*stat)(libzip_encoder_t *encoder, uint32_t *uncompressed_size, uint32_t *compressed_size, uint32_t *checksum);
    void (*free)(libzip_encoder_t *encoder);
};

/* ===== libzip_store_encoder ===== */

static bool libzip_store_encoder_write(libzip_encoder_t *encoder)
{
    bool success = false;
    size_t length;
    
    raise_if_fail(NULL != encoder);
    raise_if_fail(NULL != encoder->next_in || 0 == encoder->avail_in);
    raise_if_fail(NULL != encoder->next_out);
    raise_if_fail(encoder->avail_out > 0);

    length = encoder->avail_in;

    if (length > encoder->avail_out)
        length = encoder->avail_out;

    if (length) {
        memmove(encoder->next_out, encoder->next_in, length);
        encoder->checksum = crc32(encoder->checksum, encoder->next_in, length);
        
        encoder->avail_in-= length;
        encoder->total_in+= length;
        encoder->next_in+= length;

        encoder->avail_out-= length;
        encoder->total_out+= length;
        encoder->next_out+= length;

    }

    debugf("%d bytes read (total: %d)", length, encoder->total_in);
    debugf("%d bytes written (total: %d)", length, encoder->total_out);
    debugf("checksum=%08x", encoder->checksum);

    success = true;

finally:
    return success;
}

static libzip_encoder_t *libzip_store_encoder_new(libzip_stream_t *stream)
{
    libzip_encoder_t *encoder = NULL;
    raise_if_fail(NULL != stream);
    
    encoder = libzip_new(libzip_encoder_t, 1);
    encoder->write = libzip_store_encoder_write;
    encoder->feature_level = LIBZIP_FEATURE_LEVEL_DEFAULT;
    encoder->checksum = crc32(0, Z_NULL, 0);
    encoder->stream = stream;

finally:
    return encoder;
}

/* ===== libzip_deflate_encoder ===== */

#if ENABLE_DEFLATE

static void libzip_deflate_encoder_init(libzip_encoder_t *encoder)
{
    if (!encoder->x.deflate.initialized) {
        int zlib_result;
        z_stream *zlib;

        zlib = &encoder->x.deflate.compressor;
        zlib->zalloc = Z_NULL;
        zlib->zfree = Z_NULL;
        zlib->opaque = Z_NULL;

        /* negative value for windowBits causes zlib _not_ to write its header */
        zlib_result = deflateInit2(zlib, Z_BEST_COMPRESSION, Z_DEFLATED, -15, 9, Z_DEFAULT_STRATEGY);

        if (Z_OK != zlib_result)
            raisef("deflateInit2: %s (%d:%s)", zlib->msg, zlib_result, zError(zlib_result));

        encoder->x.deflate.initialized = (Z_OK == zlib_result);
    }

finally:
    return;
}

static bool libzip_deflate_encoder_write(libzip_encoder_t *encoder)
{
    bool success = false;
    int zlib_result;
    z_stream *zlib;
    size_t length;

    raise_if_fail(NULL != encoder);
    raise_if_fail(NULL != encoder->next_in || 0 == encoder->avail_in);
    raise_if_fail(NULL != encoder->next_out);
    raise_if_fail(encoder->avail_out > 0);

    libzip_deflate_encoder_init(encoder);
    raise_if_fail(encoder->x.deflate.initialized);

    zlib = &encoder->x.deflate.compressor;

    zlib->next_out = (void*) encoder->next_out;
    zlib->next_in = (void*) encoder->next_in;
    zlib->avail_out = encoder->avail_out;
    zlib->avail_in = encoder->avail_in;

    zlib_result = deflate(zlib, encoder->avail_in > 0 ? Z_NO_FLUSH : Z_FINISH);

    if (Z_OK != zlib_result && Z_STREAM_END != zlib_result)
        raisef("deflate failed: %s (%d)", zError(zlib_result), zlib_result);

    length = (encoder->avail_in - zlib->avail_in);

    if (length) {
        encoder->checksum = crc32(encoder->checksum, encoder->next_in, length);
        encoder->avail_in-= length;
        encoder->total_in+= length;
        encoder->next_in+= length;
    }

    length = (encoder->avail_out - zlib->avail_out);

    encoder->avail_out-= length;
    encoder->total_out+= length;
    encoder->next_out+= length;

    debugf("%d bytes read (total: %d)", length, encoder->total_in);
    debugf("%d bytes written (total: %d)", length, encoder->total_out);
    debugf("checksum=%08x", encoder->checksum);

    success = true;

finally:
    return success;
}

static libzip_encoder_t* libzip_deflate_encoder_new(libzip_stream_t *stream)
{
    libzip_encoder_t *encoder = NULL;
    raise_if_fail(NULL != stream);

    encoder = libzip_new(libzip_encoder_t, 1);
    encoder->write = libzip_deflate_encoder_write;
    encoder->feature_level = LIBZIP_FEATURE_LEVEL_DEFLATE;
    encoder->flags = LIBZIP_FLAG_DEFLATE_MAXIMUM;
    encoder->checksum = crc32(0, Z_NULL, 0);
    encoder->stream = stream;
    
finally:
    return encoder;
}

#endif

/* ===== libzip_ppmd_encoder ===== */

#if ENABLE_PPMD

int 
libzip_ppmd_encoder_getc(libzip_encoder_t *encoder __attribute__((unused)))
{
    errorf("Not implemented: %s", __FUNCTION__);
    return -1;
}

int
libzip_ppmd_encoder_putc(libzip_encoder_t *encoder __attribute__((unused)), 
                         int c __attribute__((unused))) 
{
    errorf("Not implemented: %s", __FUNCTION__);
    return -1;
}

#endif

/* ===== libzip_winzip_aes_encoder ===== */

static bool libzip_default_encoder_stat(libzip_encoder_t *encoder, uint32_t *uncompressed_size,
                                        uint32_t *compressed_size, uint32_t *checksum)
{
    bool success = false;

    raise_if_fail(NULL != encoder);

    if (uncompressed_size)
        *uncompressed_size = encoder->total_in;
    if (compressed_size)
        *compressed_size = encoder->total_out;
    if (checksum)
        *checksum = encoder->checksum;

    success = true;

finally:
    return success;
}

static libzip_encoder_t* libzip_default_encoder_new(libzip_stream_t *stream, libzip_compression_t compression)
{
    libzip_encoder_t* encoder = NULL;

    raise_if_fail(NULL != stream);

    switch(compression)
    {
        case LIBZIP_COMPRESSION_STORE:
            encoder = libzip_store_encoder_new(stream);
            break;
            
#if ENABLE_DEFLATE
        case LIBZIP_COMPRESSION_DEFLATE:
            encoder = libzip_deflate_encoder_new(stream);
            break;
#endif

        default:
            raisef("Unsupported compression method %d", compression);
            break;
    }

finally:
    return encoder;
}

/* ===== libzip_winzip_aes_encoder ===== */

static bool libzip_aes_encoder_stat(libzip_encoder_t *encoder, uint32_t *uncompressed_size,
                                    uint32_t *compressed_size, uint32_t *checksum)
{
    bool success = false;

    raise_if_fail(NULL != encoder);

    if (encoder->stream) {
        libzip_stream_close(encoder->stream);
        encoder->stream = NULL;
    }

    if (checksum)
        *checksum = 0;

    success = libzip_default_encoder_stat(encoder, uncompressed_size, compressed_size, NULL);

finally:
    return success;
}

static libzip_encoder_t* libzip_winzip_aes_encoder_new(libzip_stream_t *stream, libzip_entry_t *entry, const char *password)
{
    libzip_encoder_t *encoder = NULL;
    libzip_stream_t *encrypted_stream = NULL;
    libzip_winzip_aes_t *crypto_params = NULL;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != entry);
    raise_if_fail(NULL != password);
    raise_if_fail('\0' != password[0]);

    crypto_params = (libzip_winzip_aes_t*) libzip_entry_find_extra(entry, LIBZIP_EXTRA_SIGNATURE_WINZIP_AES);
    raise_if_fail(NULL != crypto_params);

    encrypted_stream = libzip_aes_stream_open(stream, crypto_params->mode, password);
    raise_if_fail(NULL != encrypted_stream);

    encoder = libzip_default_encoder_new(encrypted_stream, crypto_params->compression);
    raise_if_fail(NULL != encoder);

    encoder->stat = libzip_aes_encoder_stat;
    encoder->flags|= LIBZIP_FLAG_ENCRYPTED;

    encoder->total_out = // XXX das muss eleganter gehen
        SALT_LENGTH(crypto_params->mode) +
        MAC_LENGTH(crypto_params->mode) +
        PWD_VER_LENGTH;

    if (encoder->feature_level < LIBZIP_FEATURE_LEVEL_ENCRYPTION)
        encoder->feature_level = LIBZIP_FEATURE_LEVEL_ENCRYPTION;

finally:
    if (encrypted_stream && !encoder)
        libzip_stream_close(encrypted_stream);

    return encoder;
}

/* ===== libzip_encoder ===== */

libzip_encoder_t* libzip_encoder_new(libzip_stream_t *stream, libzip_entry_t *entry, const char *password)
{
    libzip_encoder_t* encoder = NULL;
    bool failure = true;

    raise_if_fail(NULL != entry);

    switch(entry->header.compression) {
        case LIBZIP_COMPRESSION_WINZIP_AES:
            encoder = libzip_winzip_aes_encoder_new(stream, entry, password);
            break;

        default:
            encoder = libzip_default_encoder_new(stream, entry->header.compression);
            break;
    }
    
    raise_if_fail(NULL != encoder);
    raise_if_fail(NULL != encoder->stream);
    raise_if_fail(NULL != encoder->write);
    
    if (entry->header.feature_level < encoder->feature_level)
        entry->header.feature_level = encoder->feature_level;

    entry->header.flags|= encoder->flags;
    failure = false;

finally:
    if (failure && encoder) {
        libzip_encoder_free(encoder);
        encoder = NULL;
    }
        
    return encoder;
}

bool libzip_encoder_write(libzip_encoder_t *encoder, const void *data, size_t length)
{
    bool success = false;
    uint8_t buffer[65536];

    raise_if_fail(NULL != encoder);
    raise_if_fail(NULL != encoder->stream);
    raise_if_fail(NULL != encoder->write);
    raise_if_fail(NULL != data || 0 == encoder->avail_in);

    encoder->next_in = data;
    encoder->avail_in = length;

    do {
        encoder->next_out = buffer;
        encoder->avail_out = sizeof(buffer);
    
        if (false == encoder->write(encoder))
            raise("Encoding failure");

        length = (sizeof(buffer) - encoder->avail_out);

        if (length != libzip_stream_write(encoder->stream, buffer, length))
            raise("Streaming failure");

    } while(encoder->avail_in);
    success = true;

finally:
    return success;
}

bool libzip_encoder_stat(libzip_encoder_t *encoder, uint32_t *uncompressed_size,
                         uint32_t *compressed_size, uint32_t *checksum)
{
    bool success = false;

    raise_if_fail(NULL != encoder);

    success = encoder->stat ?
        encoder->stat(encoder, uncompressed_size, compressed_size, checksum) :
        libzip_default_encoder_stat(encoder, uncompressed_size, compressed_size, checksum);
    
finally:
    return success;
}

void libzip_encoder_free(libzip_encoder_t *encoder)
{
    if (encoder) {
        if (encoder->free)
            encoder->free(encoder);
            
        free(encoder);
    }
}

/* vim: set sw=4 sta et: */
