/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "libzip-private.h"

#include <string.h>
#include <stdlib.h>
#include <zlib.h>

/* ===== libzip_entry_t ===== */

libzip_entry_t* libzip_entry_new(const char *filename, const time_t *timestamp)
{
    libzip_entry_t *entry = NULL;
    time_t timestamp_now;

    raise_if_fail(NULL != filename);

    if (NULL == timestamp) {
        timestamp_now = time(NULL);
        timestamp = &timestamp_now;
    }

    entry = libzip_new(libzip_entry_t, 1);
    entry->filename = strdup(filename);

    entry->header.signature = LIBZIP_SIGNATURE_LOCAL_HEADER;
    entry->header.feature_level = LIBZIP_FEATURE_LEVEL_DEFAULT;
    entry->header.flags = LIBZIP_FLAG_DESCRIPTOR_NEEDED;
    entry->header.compression = LIBZIP_COMPRESSION_STORE;
    entry->header.last_modified = libzip_dos_date_time(timestamp);
    entry->header.checksum = 0; // later
    entry->header.compressed_size = 0; // later
    entry->header.uncompressed_size = 0; // later
    entry->header.filename_length = strlen(filename);
    entry->header.extra_length = 0;
   
finally:
    return entry;
}

void libzip_entry_add_extra(libzip_entry_t *entry, const libzip_extra_t *extra)
{
    size_t length;
    
    return_if_fail(NULL != entry);
    return_if_fail(NULL != extra);

    length = sizeof(libzip_extra_t) + extra->length;
    entry->extra = realloc(entry->extra, entry->header.extra_length + length);
    memcpy(entry->extra + entry->header.extra_length, extra, length);
    entry->header.extra_length+= length;
}

const libzip_extra_t* libzip_entry_find_extra(const libzip_entry_t *entry, uint16_t signature)
{
    if (entry->extra) 
    {
	    const uint8_t *head = entry->extra;
        const uint8_t *tail = entry->extra + entry->header.extra_length;

        while(head < tail) 
        {
			// Be careful not to do a misaligned access here, as we're stepping through memory 
			// possibly an odd number of bytes at a time
			uint16_t head_sig = head[1];
			head_sig <<= 8;
			head_sig |= head[0];
			
            if(signature == head_sig) 
            {
				// Copy this memory to ensure that it is well aligned and let the caller
				// deal with deletion
				libzip_extra_t *result;
				
				uint16_t head_len = head[3];				
				head_len <<= 8;
				head_len |= head[2];
				head_len += 4;

				result = (libzip_extra_t *)calloc(head_len, 1);
				memcpy(result, head, head_len);
				
                return result;
            }
            else 
            {
				// Skip to next header - length of block + length of header (4)
				uint16_t head_len = head[3];
				head_len <<= 8;
				head_len |= head[2];

             	head += (head_len + 4);
            }
        }
    }

    return NULL;
}

void libzip_entry_free(libzip_entry_t *entry)
{
    if (entry) {
        if (entry->filename)
            free(entry->filename);
        if (entry->extra)
            free(entry->extra);

        free(entry);
    }
}

const char* libzip_entry_get_filename(const libzip_entry_t *entry)
{
    return_val_if_fail(NULL != entry, NULL);
    return entry->filename;
}

libzip_compression_t libzip_entry_get_compression(const libzip_entry_t *entry)
{
    return_val_if_fail(NULL != entry, 0);
    return entry->header.compression;
}

time_t libzip_entry_get_timestamp(const libzip_entry_t *entry)
{
    return_val_if_fail(NULL != entry, 0);
    return libzip_timestamp(entry->header.last_modified);
}

size_t libzip_entry_get_length(const libzip_entry_t *entry)
{
    return_val_if_fail(NULL != entry, 0);
    return entry->header.uncompressed_size;
}

/* vim: set sw=4 sta et: */
