/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "libzip-private.h"
#include "fileencrypt/fileenc.h"

#include <stdlib.h>

/* ===== private structures ===== */

typedef enum {
    LIBZIP_STREAM_MODE_READING = (1 << 0),
    LIBZIP_STREAM_MODE_WRITING = (1 << 1)
} libzip_stream_mode_t;

struct libzip_stream_t {
    libzip_stream_mode_t mode;

    union {
        FILE *stdio;

        struct {
            libzip_stream_t *stream;
            libzip_aes_mode_t mode;
            fcrypt_ctx cipher;

            bool header_needed;
            char *password;

            uint8_t salt[SALT_LENGTH(LIBZIP_AES_MODE_256)];
            uint8_t hash[PWD_VER_LENGTH];
        } aes;
    } x;

    long (*tell)(libzip_stream_t *stream);
    bool (*seek)(libzip_stream_t *stream, long offset, int whence);

    size_t (*read)(libzip_stream_t *stream, void *data, size_t length);
    size_t (*write)(libzip_stream_t *stream, const void *data, size_t length);
    bool (*flush)(libzip_stream_t *stream);
    bool (*close)(libzip_stream_t *stream);
};

/* ===== libzip_stdio_stream ===== */

static bool libzip_stdio_stream_seek(libzip_stream_t* stream, long offset, int whence)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.stdio);
    success = (0 == fseek(stream->x.stdio, offset, whence));

finally:
    return success;
}

static long libzip_stdio_stream_tell(libzip_stream_t* stream)
{
    long position = -1;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.stdio);
    position = ftell(stream->x.stdio);

finally:
    return position;
}

static size_t libzip_stdio_stream_read(libzip_stream_t* stream, void *data, size_t length)
{
    size_t bytes_read = 0;

    raise_if_fail(NULL != data);
    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.stdio);
    bytes_read = fread(data, 1, length, stream->x.stdio);

finally:
    debugf("bytes_read: %d", bytes_read);
    return bytes_read;
}

static size_t libzip_stdio_stream_write(libzip_stream_t* stream, const void *data, size_t length)
{
    size_t bytes_written = 0;

    raise_if_fail(NULL != data);
    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.stdio);
    bytes_written+= fwrite(data, 1, length, stream->x.stdio);

finally:
    debugf("bytes_written: %d", bytes_written);
    return bytes_written;
}

static bool libzip_stdio_stream_flush(libzip_stream_t* stream)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.stdio);
    success = (0 == fflush(stream->x.stdio));

finally:
    return success;
}

static bool libzip_stdio_stream_close(libzip_stream_t* stream)
{
    size_t success = 0;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.stdio);
    success = (0 == fclose(stream->x.stdio));
    stream->x.stdio = NULL;

finally:
    return success;
}

libzip_stream_t* libzip_stdio_stream_open(FILE *stdio_stream)
{
    libzip_stream_t *stream = NULL;

    raise_if_fail(NULL != stdio_stream);
    stream = libzip_new(libzip_stream_t, 1);

    stream->x.stdio = stdio_stream;
    stream->tell = libzip_stdio_stream_tell;
    stream->seek = libzip_stdio_stream_seek;
    stream->read = libzip_stdio_stream_read;
    stream->write = libzip_stdio_stream_write;
    stream->flush = libzip_stdio_stream_flush;
    stream->close = NULL; // as we do not own the stream


finally:
    return stream;
}

/* ===== libzip_aes_stream ===== */

static void dump_bytes(const char *prefix, const void *data, size_t length)
{
    const uint8_t *head, *tail;
    return_if_fail(NULL != data);

    head = data; 
    tail = head + length;

    printf("%s:", prefix ? prefix : "bytes");
    
    while(head < tail)
        printf(" %02x", *head++);

    printf("\n");
}

static bool libzip_aes_stream_read_header(libzip_stream_t* stream)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.aes.stream);
    raise_if_fail(0 == (stream->mode & LIBZIP_STREAM_MODE_WRITING));

    if (stream->x.aes.header_needed) {
        const size_t salt_length = SALT_LENGTH(stream->x.aes.mode);
        const size_t hash_length = PWD_VER_LENGTH;
        uint8_t stored_hash[PWD_VER_LENGTH];
        int fcrypt_result;

        if (salt_length != libzip_stream_read(stream->x.aes.stream, stream->x.aes.salt, salt_length))
            raise("Cannot read AES salt.");
        if (hash_length != libzip_stream_read(stream->x.aes.stream, stored_hash, hash_length))
            raise("Cannot read AES hash.");

        fcrypt_result = fcrypt_init(stream->x.aes.mode,
            (uint8_t*) stream->x.aes.password, strlen(stream->x.aes.password), 
            stream->x.aes.salt, stream->x.aes.hash, &stream->x.aes.cipher);
        debugf("fcrypt_init: %d", fcrypt_result);
        raise_if_fail(0 == fcrypt_result);

        if (libzip_debug_level & 4) {
            dump_bytes("stream->x.aes.salt", stream->x.aes.salt, salt_length);
            dump_bytes("stream->x.aes.hash", stream->x.aes.hash, hash_length);
            dump_bytes("stored_hash", stored_hash, hash_length);
        }

        if (memcmp(stored_hash, stream->x.aes.hash, hash_length))
            raise("Password does not match.");

        stream->x.aes.header_needed = false;
        success = true;
    } else {
        success = true;
    }

finally:
    return success;
}

static size_t libzip_aes_stream_read(libzip_stream_t* stream, void *data, size_t length)
{
    size_t bytes_read = 0;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.aes.stream);
    raise_if_fail(0 == (stream->mode & LIBZIP_STREAM_MODE_WRITING));

    if (false == libzip_aes_stream_read_header(stream))
        raise("Cannot read encryption header.");

    if (length) {
        bytes_read = libzip_stream_read(stream->x.aes.stream, data, length);
        raise_if_fail(bytes_read <= length);

        debugf("decrypting %d bytes", bytes_read);
        fcrypt_decrypt(data, bytes_read, &stream->x.aes.cipher);
    }

finally:
    debugf("bytes_read: %d", bytes_read);
    return bytes_read;
}

static bool libzip_aes_stream_write_header(libzip_stream_t* stream)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.aes.stream);
    raise_if_fail(NULL != stream->x.aes.password);
    raise_if_fail(0 == (stream->mode & LIBZIP_STREAM_MODE_READING));

    if (stream->x.aes.header_needed) {
        const size_t salt_length = SALT_LENGTH(stream->x.aes.mode);
        const size_t hash_length = PWD_VER_LENGTH;
        int fcrypt_result;

        if (libzip_debug_level & 8) {
	        size_t i;
            warn("Using fixed AES SALT for testing!");
            
            for(i = 0; i < SALT_LENGTH(stream->x.aes.mode); ++i)
                stream->x.aes.salt[i] = 'A' + i;
        } else {
            libzip_random(stream->x.aes.salt, SALT_LENGTH(stream->x.aes.mode));
        }

        fcrypt_result = fcrypt_init(stream->x.aes.mode,
            (uint8_t*) stream->x.aes.password, strlen(stream->x.aes.password), 
            stream->x.aes.salt, stream->x.aes.hash, &stream->x.aes.cipher);
        debugf("fcrypt_init: %d", fcrypt_result);
        raise_if_fail(0 == fcrypt_result);

        if (libzip_debug_level & 4) {
            dump_bytes("stream->x.aes.salt", stream->x.aes.salt, salt_length);
            dump_bytes("stream->x.aes.hash", stream->x.aes.hash, hash_length);
        }

        if (salt_length != libzip_stream_write(stream->x.aes.stream, stream->x.aes.salt, salt_length))
            raise("Cannot write AES salt.");
        if (hash_length != libzip_stream_write(stream->x.aes.stream, stream->x.aes.hash, hash_length))
            raise("Cannot write AES hash.");

        stream->x.aes.header_needed = false;
        success = true;
    } else {
        success = true;
    }

finally:
    return success;
}

static size_t libzip_aes_stream_write(libzip_stream_t* stream, const void *data, size_t length)
{
    size_t bytes_written = 0;

    raise_if_fail(NULL != data);
    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.aes.stream);
    raise_if_fail(0 == (stream->mode & LIBZIP_STREAM_MODE_READING));

    if (false == libzip_aes_stream_write_header(stream))
        raise("Cannot write encryption header.");

    if (length) {
        uint8_t buffer[65536];

        if (length > sizeof(buffer))
            length = sizeof(buffer);

        memcpy(buffer, data, length);
        debugf("encrypting %d bytes", length);
        fcrypt_encrypt(buffer, length, &stream->x.aes.cipher);
        bytes_written+= libzip_stream_write(stream->x.aes.stream, buffer, length);
    }

finally:
    debugf("bytes_written: %d", bytes_written);
    return bytes_written;
}

static bool libzip_aes_stream_flush(libzip_stream_t* stream)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.aes.stream);
//    raise_if_fail(0 == (stream->mode & LIBZIP_STREAM_MODE_READING));
    
    success = libzip_stream_flush(stream->x.aes.stream);

finally:
    return success;
}

static bool libzip_aes_stream_close(libzip_stream_t* stream)
{
    size_t success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->x.aes.stream);

    if ((stream->mode & LIBZIP_STREAM_MODE_WRITING)) {
        uint8_t mac[MAC_LENGTH(LIBZIP_AES_MODE_256)];
        size_t mac_length;

        raise_if_fail(sizeof(mac) >= MAC_LENGTH(stream->x.aes.mode));
    
        if (false == libzip_aes_stream_write_header(stream))
            raise("Cannot write encryption header.");

        if (false == libzip_stream_flush(stream->x.aes.stream))
            raise("Final flushing of encrypted stream failed.");

        mac_length = fcrypt_end(mac, &stream->x.aes.cipher);
        debugf("fcrypt_end: mac_length=%d", mac_length);
        raise_if_fail(sizeof(mac) >= mac_length);

        if (libzip_debug_level & 4) {
            dump_bytes("mac", mac, mac_length);
        }
        
        if (mac_length != libzip_stream_write(stream->x.aes.stream, mac, mac_length))
            raise("Cannot write AES checksum.");

    }
    
    if (stream->x.aes.password) {
        free(stream->x.aes.password);
        stream->x.aes.password = NULL;
    }

    stream->x.aes.stream = NULL;
    success = true;

finally:
    return success;
}

libzip_stream_t* libzip_aes_stream_open(libzip_stream_t* wrapped, libzip_aes_mode_t mode, const char *password)
{
    libzip_stream_t *stream = NULL;
    bool failure = true;

    raise_if_fail(NULL != wrapped);
    raise_if_fail(NULL != password);
    raise_if_fail('\0' != password[0]);
    
    raise_if_fail(mode >= LIBZIP_AES_MODE_128);
    raise_if_fail(mode <= LIBZIP_AES_MODE_256);
    
    raise_if_fail(sizeof(stream->x.aes.hash) >= PWD_VER_LENGTH);
    raise_if_fail(sizeof(stream->x.aes.salt) >= SALT_LENGTH(mode));

    stream = libzip_new(libzip_stream_t, 1);

    stream->x.aes.mode = mode;
    stream->x.aes.stream = wrapped;
    stream->x.aes.header_needed = true;
    stream->x.aes.password = strdup(password);

    stream->read = libzip_aes_stream_read;
    stream->write = libzip_aes_stream_write;
    stream->flush = libzip_aes_stream_flush;
    stream->close = libzip_aes_stream_close;

    failure = false;

finally:
    if (failure && stream) {
        libzip_stream_close(stream);
        stream = NULL;
    }

    return stream;
}

/* ===== libzip_stream ===== */

libzip_stream_t* libzip_stream_open(const char *filename, const char *mode)
{
    FILE *stdio_stream = NULL;
    libzip_stream_t *stream = NULL;

    raise_if_fail(NULL != filename);
    raise_if_fail(NULL != mode);

    if ((stdio_stream = fopen(filename, mode)) &&
        (stream = libzip_stdio_stream_open(stdio_stream)))
        stream->close = libzip_stdio_stream_close;

finally:
    return stream;
}

long libzip_stream_tell(libzip_stream_t *stream)
{
    long position = -1;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->tell);
    position = stream->tell(stream);

finally:
    return position;
}

bool libzip_stream_seek(libzip_stream_t *stream, long offset, int whence)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->seek);
    success = stream->seek(stream, offset, whence);

finally:
    return success;
}

size_t libzip_stream_read(libzip_stream_t *stream, void *data, size_t length)
{
    size_t bytes_read = 0;

    raise_if_fail(NULL != data);
    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->read);

    stream->mode|= LIBZIP_STREAM_MODE_READING;
    bytes_read = stream->read(stream, data, length);

finally:
    debugf("bytes_read: %d", bytes_read);
    return bytes_read;
}

size_t libzip_stream_write(libzip_stream_t *stream, const void *data, size_t length)
{
    size_t bytes_written = 0;

    raise_if_fail(NULL != data);
    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->write);

    stream->mode|= LIBZIP_STREAM_MODE_WRITING;
    bytes_written = stream->write(stream, data, length);

finally:
    debugf("bytes_written: %d", bytes_written);
    return bytes_written;
}

bool libzip_stream_flush(libzip_stream_t *stream)
{
    bool success = false;

    raise_if_fail(NULL != stream);
    raise_if_fail(NULL != stream->flush);
    success = stream->flush(stream);

finally:
    return success;
}

bool libzip_stream_close(libzip_stream_t *stream)
{
    bool success = false;

    if (stream) {
        success =
            (true == libzip_stream_flush(stream)) &&
            (NULL == stream->close || stream->close(stream));

        free(stream);
    }

    return success;
}

/* vim: set sw=4 sta et: */
