#include "libzip.h"

#include <getopt.h>
#include <stdlib.h>

#define raise(Message) do { \
    fprintf(stderr, "[mkzip] ERROR: %s (%s, line %d): " Message "\n", \
            __FUNCTION__, __FILE__, __LINE__); goto finally; \
} while(0)

#define raisef(Format, ...) do { \
    fprintf(stderr, "[mkzip] ERROR: %s (%s, line %d): " Format "\n", \
            __FUNCTION__, __FILE__, __LINE__, __VA_ARGS__); goto finally; \
} while(0)

static void print_usage()
{
    printf(
        "Usage: mkzip [OPTIONS] FILES\n"
        "Creates PKZIP compatible file archives containing the files specified.\n"
        "\n"
        "Options:\n"
        "  -a LEVEL      AES encryption level (0: none, 1: 128, 2: 192, 3: 256 bit)\n"
        "  -c METHOD     compression method to use (0: store, 8: deflate)\n"
        "  -d BITMASK    debug messages to show\n"
        "  -o FILENAME   write archive to file, instead of standard output\n"
        "  -p PASSWORD   password to use\n"
        "\n"
        "Copyright 2005 Mathias Hasselmann <mathias@taschenorakel.de>\n"
        "Bug reports: https://taschenorakel.de/bugzilla/enter_bug.cgi?product=libzip\n");
}

int main(int argc, char* argv[]) {
    libzip_aes_mode_t winzip_aes_mode = LIBZIP_AES_MODE_NONE;
    libzip_compression_t compression = LIBZIP_COMPRESSION_STORE;
    const char *password = NULL;
    const char *output = NULL;

    int i, rc = 1;

    while(-1 != (i = getopt(argc, argv, "a:c:d:o:p:h"))) {
        switch(i) {
            case 'a':
                winzip_aes_mode = atoi(optarg);
                break;

            case 'c':
                compression = atoi(optarg);
                break;

            case 'd':
                libzip_set_debug_level(atoi(optarg));
                break;

            case 'o':
                output= optarg;
                break;

            case 'p':
                password = optarg;
                break;

            case 'h':
                print_usage();
                return 0;

            default:
                fprintf(stderr, "Usage error. Call 'mkzip -h' for help.\n");
                return 1;
        }
    }

    libzip_stream_t *stream = NULL;
    libzip_writer_t *writer = NULL;

    if (output) {
        stream = libzip_stream_open(output, "wb");
    } else {
        stream = libzip_stdio_stream_open(stdout);
    }

    if (NULL == output)
        raise("Cannot open stream. Aborting.");
    if (NULL == (writer = libzip_writer_create(stream)))
        raise("Cannot create ZIP writer. Aborting.");
    
    libzip_writer_set_compression(writer, compression);
    libzip_writer_set_winzip_aes_mode(writer, winzip_aes_mode);

    if (password)
        libzip_writer_set_password(writer, password);

    for(i = optind; i < argc; ++i) {
        libzip_entry_t *entry = libzip_writer_add_file(writer, NULL, argv[i]);

        if (NULL == entry)
            raisef("Cannot store '%s'. Aborting.", argv[i]);
    }

    rc = 0;

finally:
    if (writer)
        libzip_writer_close(writer);
    if (stream)
        libzip_stream_close(stream);

    return rc;
}

/* vim: set sw=4 sta et: */
